/* ---- file bigcmp.c ---- */

/* Comparisons of factored values larger than an int
   by C.B. Falconer.
   Bug reports to: <mailto:cbfalconer@worldnet.att.net>
   put in public domain.  Attributions appreciated.
   The function of interest is "decide".  The rest is
   simply testing convenience code.
   This can be found at:
       <http://cbfalconer.home.att.net/download/bigcmp.zip>
*/

/* The purpose of this is to take two number pairs
     A, B  and  C, D
   and determine if ((A * B) > (C * D))

   without creating any overflows, etc.  For this
   demonstration we will assume A, B, C, D are all
   greater or equal to zero.

   First we find the minimum of A and C, and of B and D.  Then
   we can express the relationship as:

       (c + x) * (d + y) < c * d
   or
       (c + x) * b < c * (b + y)

   taking only the first possibility, it becomes:

       c * d + c * y + x * d + x * y < c * d
   or
       y * (c + x) + x * y < 0

   which is obviously false, since all values are positive.  That
   leaves the second expression to worry about, which reduces to:

       x * b < c * y again with all values positive.  This is the
   original problem with smaller values.

   (we have to watch out for zeroes arising, which leads to a
   very simple case)
*/

#include <stdio.h>
#include <stdlib.h> /* strtol */
#include <errno.h>  /* errno, ERANGE */
#include <limits.h> /* INT_MAX */

/* ------------------------ */

/* Exchange two integers */
static void xchange(int *a, int *b)
{
   int t;

   t = *a; *a = *b; *b = t;
} /* xchange */

/* ------------------------ */

/* a return value of 1 means a*b  < c*d */
/*                  -1 means a*b  > c*d */
/*                   0 means a*b == c*d */
int decide(int a, int b, int c, int d)
{
   int comp;  /* result of partial comparison */
   int flip;  /* Whether to invert the result */

#define negate(n) do if (n > 0) {n = -n; flip *= -1;} while (0)

#if DEBUG
      printf("\ndecide(%d,%d,%d,%d)\n", a, b, c, d);
#endif

   flip = 1;

   /* if there is any range limit, -ve can cover it */
   /* detect incoming zeroes here. */
   if (a && b) {  /* no immediate decision */
     negate(a); negate(b);
   }
   else if (!c || !d) return 0; /* both sides 0 */
   else {                       /* only a*b == 0 */
      negate(c); negate(d);
      return flip;
   }

   /* a * b != 0, flip is sign of product a*b */
   comp = flip;
   if (!c || !d) return -flip;
   else {
      negate(c); negate(d);
   }
   /* Now all values are in the range INT_MIN .. -1 */
   if (flip < 0) {
      /* a * b has different sign than c * d) */
      return -comp;
   }
   flip = comp;  /* i.e. sign of a*b (BUGFIX) */

   /* This is the point - compare without overflows */
   do {
#if DEBUG
      printf("decide(%d,%d,%d,%d)   flip = %d\n",
              a, b, c, d, flip);
#endif
      /* speed up the convergences */
      if (a > b) xchange(&a, &b);
      if (c > d) xchange(&c, &d);

      if      (a == c)    comp = 0;
      else if (a < c)     comp = 1;
      else if (a > c) {
         xchange(&a, &c); xchange(&b, &d);
         flip = -flip;    comp = 1;
      }
#if DEBUG
      printf("now   (%d,%d,%d,%d)   flip = %d comp = %d\n",
              a, b, c, d, flip, comp);
#endif
      /* now a <= c. Remember, all negative */
      if (b == d)
         return comp * -flip;
      else if (b < d)
         return -flip;

      /* now this cannot create any zeroes */
      a = a-c; d = d-b; /* go around again */
   } while (1);
} /* decide */

/* ------------------------ */

static void help(char *msg1, char *msg2)
{
   if (msg1) fputs(msg1, stderr);
   if (msg2) fputs(msg2, stderr);
   fputs("\nusage: bigcmp a b c d\n"
         "where a, b, c, d are positive integer values\n"
         "that do not exceed the capacity of an int\n"
         "or\n"
         "       bigcmp datafile (for regression testing)\n",
         stderr);
   exit(EXIT_FAILURE);
} /* help */

/* ------------------------ */

static int getnumber(char *param)
{
   long  value;
   char *errp;

   value = strtol(param, &errp, 10);
   if (errp == param) help("no value: ", param);
   if ((ERANGE == errno) ||
       (value > INT_MAX) || (value < INT_MIN))
      help("too big: ", param);
   return value;
} /* getnumber */

/* ------------------------ */

int testone(int a, int b, int c, int d)
{
   double axb, cxd;
   int    cmp;
   char   ch;

   axb = (double)a * b;
   cxd = (double)c * d;

   cmp = decide(a, b, c, d);
   switch (cmp) {
case -1: ch = '>'; break;
case 0:  ch = '='; break;
case 1:  ch = '<'; break;
default: printf("This can't happen\n");
         exit(EXIT_FAILURE);
   }
   printf("decide(%d, %d, %d, %d) returns %d, "
          "i.e. %.0f %c %.0f\n",
          a, b, c, d,
          cmp, axb, ch, cxd);
   return cmp;
} /* testone */

/* ------------------------ */

int main(int argc, char **argv)
{
   int    a, b, c, d;
   int    cmp;
   FILE  *fp;
   int    items, fails;

   if (2 == argc) {
      if (!(fp = fopen(argv[1], "r"))) {
         help("No data file: ", argv[1]);
      }
      else { /* use the data file to run regression tests */
         items = fails = 0;
         while (5 == fscanf(fp, "%d %d %d %d %d",
                            &a, &b, &c, &d, &cmp)) {
            items++;
            if (cmp != testone(a, b, c, d)) {
               fails++;
               printf("Failure #%d test %d\n\n", fails, items);
            }
         }
         if (!feof(fp))
            help("Bad data format in ", argv[1]);
         else {
            printf("%d tests, %d failures\n", items, fails);
            if (fails) return EXIT_FAILURE;
         }
      }
   }
   else if (argc < 5) help(NULL, NULL); /* and exit */
   else {
      a = getnumber(argv[1]);
      b = getnumber(argv[2]);
      c = getnumber(argv[3]);
      d = getnumber(argv[4]);

      cmp = testone(a, b, c, d);
   }
   return 0;
} /* main */
/* ----- EOF bigcmp.c ----- */
